//
//  TestTVC.swift
//  EZCloudKit
//
//  Created by Tim Roadley on 9/10/2015.
//  Copyright © 2015 Tim Roadley. All rights reserved.
//

import UIKit
import CoreData
import CloudKit

class TestTVC: CDTableViewController {
    
    // MARK: - CELL CONFIGURATION
    
    override func configureCell(cell: UITableViewCell, atIndexPath indexPath: NSIndexPath) {
    
        if let test = self.frc.objectAtIndexPath(indexPath) as? Test {
        
            if let textLabel = cell.textLabel,
                      device = test.device {
                textLabel.text = device
            }
            if let detailTextLabel = cell.detailTextLabel,
                          modified = test.modified,
                         someValue = test.someValue {
                detailTextLabel.text = "\(modified) \(someValue)"
            }
        }
    }
    
    // MARK: - INITIALIZATION
    required init?(coder aDecoder: NSCoder) {
        super.init(coder: aDecoder)
    
        // CDTableViewController subclass customization
        self.entity = "Test"
        self.sort = [NSSortDescriptor(key: "someValue", ascending: true)]
        self.filter = NSPredicate(format: "group == %@ AND groupKey == %@ AND cacheStatus != %i", CDCloudSync.shared.group(), CDCloudSync.shared.groupKey(), DELETING)
        self.fetchBatchSize = 25
    }
    
    // MARK: - VIEW
    override func viewDidLoad() {
        super.viewDidLoad()
        self.navigationItem.title = "EZCloudKit"
        self.performFetch()
    
        // Enable CloudSync Swipe Down to Refresh
        NSNotificationCenter.defaultCenter().addObserver(self, selector: "syncMessage:", name: "syncMessage", object: nil)
        self.refreshControl = UIRefreshControl()
        if let refreshControl = self.refreshControl {
            refreshControl.addTarget(self, action: Selector("sync"), forControlEvents: UIControlEvents.ValueChanged)
        }
    }
    override func viewDidAppear(animated: Bool) {
        super.viewDidAppear(animated)
        CDDeduplicator.deDuplicateEntityWithName("Test", uniqueAttributeName: "someValue", backgroundMoc: CDHelper.shared.importContext)
    }
    
    // MARK: - INTERACTION
    @IBAction func add (sender: AnyObject) {
    
        if let object = NSEntityDescription.insertNewObjectForEntityForName("Test", inManagedObjectContext: CDHelper.shared.context) as? Test {
            object.device = UIDevice().name
            object.modified = NSDate()
            object.someValue = "Test: \(NSUUID().UUIDString)"
            CDCloudSync.shared.setCacheStatusForObject(object, requestedStatus: NEW)
            CDCloudSync.addGroupInfo(object)
            CDHelper.saveSharedContext()
        }
    }
    
    // MARK: - DATA SOURCE: UITableView
    override func tableView(tableView: UITableView, commitEditingStyle editingStyle: UITableViewCellEditingStyle, forRowAtIndexPath indexPath: NSIndexPath) {
    
        if (editingStyle == UITableViewCellEditingStyle.Delete) {
    
            if let object = self.frc.objectAtIndexPath(indexPath) as? NSManagedObject {
                CDCloudSync.shared.setCacheStatusForObject(object, requestedStatus: DELETING)
                CDHelper.saveSharedContext()
            }
        }
    }
    
    // MARK: - CLOUDSYNC
    func groupChanged () {
        let predicate = NSPredicate(format: "group == %@ AND groupKey == %@ AND cacheStatus != %i", CDCloudSync.shared.group(), CDCloudSync.shared.groupKey(), DELETING)
        self.filter = predicate
        self.frc.fetchRequest.predicate = predicate
        self.performFetch()
    }
    func sync () {
        let defaultContainer = CKContainer.defaultContainer()
        let publicDatabase = defaultContainer.publicCloudDatabase
        CDCloudSync.shared.syncEntities(["Test"], database: publicDatabase, context: CDHelper.shared.importContext)
    }
    func syncMessage (note:NSNotification) {
        
        if let title = note.object as? String {
            self.navigationItem.title = title
        
            if let refreshControl = self.refreshControl {
                refreshControl.endRefreshing()
                self.performFetch()
            }
        }
    }
}