//
//  TestTVC.swift
//  EZiCloud
//
//  Created by Tim Roadley on 30/11/2015.
//  Copyright © 2015 Tim Roadley. All rights reserved.
//

import UIKit
import CoreData

class TestTVC: CDTableViewController {
    
    // MARK: - CELL CONFIGURATION
    
    override func configureCell(cell: UITableViewCell, atIndexPath indexPath: NSIndexPath) {
    
        if let test = self.frc.objectAtIndexPath(indexPath) as? Test {
        
            if let textLabel = cell.textLabel,
                      device = test.device {
                textLabel.text = device
            }
            if let detailTextLabel = cell.detailTextLabel,
                          modified = test.modified,
                         someValue = test.someValue {
                detailTextLabel.text = "\(modified) \(someValue)"
            }
        }
    }
    
    // MARK: - INITIALIZATION
    required init?(coder aDecoder: NSCoder) {
        super.init(coder: aDecoder)
    
        // CDTableViewController subclass customization
        self.entity = "Test"
        self.sort = [NSSortDescriptor(key: "modified", ascending: false)]
        self.fetchBatchSize = 25
    }
    
    // MARK: - VIEW
    override func viewDidLoad() {
        super.viewDidLoad()
        self.navigationItem.title = "EZiCloud"
        self.performFetch()
    }
    override func viewDidAppear(animated: Bool) {
        super.viewDidAppear(animated)
        CDDeduplicator.deDuplicateEntityWithName("Test", uniqueAttributeName: "someValue", backgroundMoc: CDHelper.shared.importContext)
    }
    
    // MARK: - INTERACTION
    @IBAction func add (sender: AnyObject) {
    
    if let object = NSEntityDescription.insertNewObjectForEntityForName("Test", inManagedObjectContext: CDHelper.shared.context) as? Test {
            object.device = UIDevice().name
            object.modified = NSDate()
            object.someValue = "Test: \(NSUUID().UUIDString)"
            CDHelper.saveSharedContext()
        }
    }
    
    // MARK: - DATA SOURCE: UITableView
    override func tableView(tableView: UITableView, commitEditingStyle editingStyle: UITableViewCellEditingStyle, forRowAtIndexPath indexPath: NSIndexPath) {
    
        if (editingStyle == UITableViewCellEditingStyle.Delete) {
    
            if let object = self.frc.objectAtIndexPath(indexPath) as? NSManagedObject {
                self.frc.managedObjectContext.deleteObject(object)
            }
            CDHelper.saveSharedContext()
        }
    }
} 