//
//  AppDelegate.swift
//  Groceries
//
//  Created by Tim Roadley on 29/09/2015.
//  Copyright © 2015 Tim Roadley. All rights reserved.
//

import UIKit
import CoreData

@UIApplicationMain
class AppDelegate: UIResponder, UIApplicationDelegate {

    var window: UIWindow?

    func demo () {
        if CDOperation.objectCountForEntity("Item", context: CDHelper.shared.context) == 0 {
            
            let moc = CDHelper.shared.importContext
            moc.performBlock {
            
                if let locationAtHome = NSEntityDescription.insertNewObjectForEntityForName("LocationAtHome", inManagedObjectContext: moc) as? LocationAtHome
                 , let locationAtShop = NSEntityDescription.insertNewObjectForEntityForName("LocationAtShop", inManagedObjectContext: moc) as? LocationAtShop
                 , let unit = NSEntityDescription.insertNewObjectForEntityForName("Unit", inManagedObjectContext: moc) as? Unit {
            
                    locationAtHome.storedIn = "Test Home Location"
                    locationAtShop.aisle = "Test Shop Location"
                    unit.name = ""
                    
                    for i in 1...100 {
                            
                        if let item = NSEntityDescription.insertNewObjectForEntityForName("Item", inManagedObjectContext: moc) as? Item {
                            item.locationAtHome = locationAtHome
                            item.locationAtShop = locationAtShop
                            item.unit = unit
                            item.name = "Test Item \(i)"
                            
                            if let photo = UIImage(named: "BigGroceries.png"), photoData = UIImagePNGRepresentation(photo)  {
        
                                if let itemPhoto = NSEntityDescription.insertNewObjectForEntityForName("Item_Photo", inManagedObjectContext: moc) as? Item_Photo {
                                    itemPhoto.data = photoData
                                    item.photo = itemPhoto
                                    CDFaulter.faultObject(itemPhoto, moc: moc)
                                }
                            }
                            
                            // Save to disk, free up memory and trigger interface refresh
                            CDFaulter.faultObject(item, moc: moc)
                        }
                        CDHelper.shared.context.performBlock {
                            NSNotificationCenter.defaultCenter().postNotificationName("SomethingChanged", object: nil)
                        }
                    }
                }
            }
        }
    }

    func application(application: UIApplication, didFinishLaunchingWithOptions launchOptions: [NSObject: AnyObject]?) -> Bool {
        // Override point for customization after application launch.
        
        return true
    }

    func applicationWillResignActive(application: UIApplication) {
        // Sent when the application is about to move from active to inactive state. This can occur for certain types of temporary interruptions (such as an incoming phone call or SMS message) or when the user quits the application and it begins the transition to the background state.
        // Use this method to pause ongoing tasks, disable timers, and throttle down OpenGL ES frame rates. Games should use this method to pause the game.
    }

    func applicationDidEnterBackground(application: UIApplication) {
        // Use this method to release shared resources, save user data, invalidate timers, and store enough application state information to restore your application to its current state in case it is terminated later.
        // If your application supports background execution, this method is called instead of applicationWillTerminate: when the user quits.
        
        CDHelper.saveSharedContext()
    }

    func applicationWillEnterForeground(application: UIApplication) {
        // Called as part of the transition from the background to the inactive state; here you can undo many of the changes made on entering the background.
    }

    func applicationDidBecomeActive(application: UIApplication) {

        if CDCloudSync.shared.group() == "Public" {
            if let window = self.window, view = window.rootViewController {            
                let alert = UIAlertController(title: "You're using a Public Shopping List.", message: "Set the same Shopping List Name and Password on two or more devices to stay synchronized.", preferredStyle: UIAlertControllerStyle.Alert)
                
                alert.addTextFieldWithConfigurationHandler({ (textField) -> Void in
                    textField.placeholder = "Shopping List Name"})
                
                alert.addTextFieldWithConfigurationHandler({ (textField) -> Void in
                    textField.placeholder = "Password"
                    textField.secureTextEntry = true})
                
                let saveAction = UIAlertAction(title: "Save", style: .Default, handler: { (action) -> Void in                
                    if let textFields = alert.textFields {
                        
                        if let shoppingListName = textFields[0].text, let password = textFields[1].text {

                            // Save custom shopping list
                            let defaults = NSUserDefaults.standardUserDefaults()
                            defaults.setObject(shoppingListName, forKey: "group")
                            defaults.setObject(password, forKey: "groupKey")
                            defaults.synchronize()
                            
                            // Reset the predicate filter on all fetched results controllers
                            NSNotificationCenter.defaultCenter().postNotificationName("GroupChanged", object: nil)
                        }
                    } else {print("\(__FUNCTION__) FAILED to get textFields from alert controller")}
                })
                alert.addAction(saveAction)
                view.presentViewController(alert, animated: true, completion: nil)
            } else {"FAILED to get window.rootViewController"}
        }
    }

    func applicationWillTerminate(application: UIApplication) {
        // Called when the application is about to terminate. Save data if appropriate. See also applicationDidEnterBackground:.
        
        CDHelper.saveSharedContext()
    }
}

