//
//  CDCloudSync.swift
//  Groceries
//
//  Created by Tim Roadley on 8/11/2015.
//  Copyright © 2015 Tim Roadley. All rights reserved.
//

import UIKit
import CoreData
import CloudKit
import SystemConfiguration

// CACHE STATUS
let NEW = 0, INSERTING = 1, RELATING = 2, CHANGED = 3, DELETING = 4, SYNCHRONIZED = 5

private let _sharedCDCloudSync = CDCloudSync()
class CDCloudSync: NSObject {
    
    // MARK: - SHARED INSTANCE
    class var shared : CDCloudSync {
        return _sharedCDCloudSync
    }
    
    // MARK: - LEGIBILITY
    func friendlyObjectName (object:NSManagedObject) -> String {
        
        if let entityName = object.entity.name, moc = object.managedObjectContext {
            if let entity = NSEntityDescription.entityForName(entityName, inManagedObjectContext: moc) {
   
                for (attribute, _) in entity.attributesByName {
    
                    if ["name","storedIn","aisle"].contains(attribute) {
    
                        if let value = object.valueForKey(attribute) as? String {
                                
                            if let cacheID = object.valueForKey("cacheID") as? String {
                                return "[\(entityName) object \"\(value)\" cacheID:\"\(cacheID)\"]"
                            } else {
                                return "[\(entityName) object \"\(value)\" cacheID:\"nil\")]"
                            }
                        }
                    }
                }
            }
            return "[\(entityName) object]"
        }
        return "[????]"
    }
    func friendlyRecordName (record:CKRecord) -> String {
        
        if let name = record.valueForKey("name") as? String {
            return "[\(record.recordType) record \"\(name)\" recordID:\"\(record.recordID.recordName)\"]"
        }
        if let storedIn = record.valueForKey("storedIn") as? String {
            return "[\(record.recordType) record \"\(storedIn)\" recordID:\"\(record.recordID.recordName)\"]"
        }
        if let aisle = record.valueForKey("aisle") as? String {
            return "[\(record.recordType) record \"\(aisle)\" recordID:\"\(record.recordID.recordName)\"]"
        }
        return "[\(record.recordType) record recordID:\"\(record.recordID.recordName)\"]"
    }        
    func friendlyRecordSummary (record:CKRecord) -> String {
        
        if let name = record.valueForKey("name") as? String {
            return "\(name)"
        }
        if let storedIn = record.valueForKey("storedIn") as? String {
            return "\(storedIn)"
        }
        if let aisle = record.valueForKey("aisle") as? String {
            return "\(aisle)"
        }
        return "Photos"
    }
    
    // MARK: - GROUP
    func group () -> String {
        let defaults = NSUserDefaults.standardUserDefaults()
        if let object = defaults.objectForKey("group") as? String {
            return object
        }
        defaults.setObject("Public", forKey: "group")
        return "Public"
    }
    func groupKey () -> String {
        let defaults = NSUserDefaults.standardUserDefaults() 
        if let object = defaults.objectForKey("groupKey") as? String {
            return object
        }
        defaults.setObject("Public", forKey: "groupKey")
        return "Public"
    }
    class func addGroupInfo(object:NSManagedObject) {
        object.setValue(CDCloudSync.shared.group(), forKey: "group")
        object.setValue(CDCloudSync.shared.groupKey(), forKey: "groupKey")
    
        CDCloudSync.shared.setCacheStatusForObject(object, requestedStatus: CHANGED)
    }
 
    // MARK: - CACHE STATUS
    func cacheStatusName (cacheStatus:Int) -> String {
        
        switch (cacheStatus) {
            
        case NEW:          return "NEW"
        case INSERTING:    return "INSERTING"
        case RELATING:     return "RELATING"
        case CHANGED:      return "CHANGED"
        case DELETING:     return "DELETING"
        case SYNCHRONIZED: return "SYNCHRONIZED"
        default:           return "UNKNOWN"
        }
    }
    func setCacheStatusForObject (object:NSManagedObject, requestedStatus:Int) {
        
        if let currentStatus = object.valueForKey("cacheStatus") as? NSNumber {
            
            if currentStatus == NEW && requestedStatus == CHANGED {
                NSLog("%@ cannot move directly from NEW to CHANGED", self.friendlyObjectName(object))
                return
            }
            
            if currentStatus == INSERTING && requestedStatus == CHANGED {
                NSLog("%@ cannot move directly from INSERTING to CHANGED", self.friendlyObjectName(object))
                return
            }
            
            let newStatus = NSNumber(integer: requestedStatus)
            object.setValue(newStatus, forKey: "cacheStatus")
            object.setValue(NSDate(), forKey: "modified")
            if let context = object.managedObjectContext {CDHelper.save(context)}
            
        } else {NSLog("ERROR getting existing cacheStatus in %@",__FUNCTION__)}
    }
    func cacheStatusForAllObjects (entities:[String], context:NSManagedObjectContext) {
        
        var new          = [NSManagedObject]()
        var inserting    = [NSManagedObject]()
        var relating     = [NSManagedObject]()
        var changed      = [NSManagedObject]()
        var deleting     = [NSManagedObject]()
        var synchronized = [NSManagedObject]()
        
        let predicateNew          = NSPredicate(format: "cacheStatus == %i", NEW)
        let predicateInserting    = NSPredicate(format: "cacheStatus == %i", INSERTING)
        let predicateRelating     = NSPredicate(format: "cacheStatus == %i", RELATING)
        let predicateChanged      = NSPredicate(format: "cacheStatus == %i", CHANGED)
        let predicateDeleting     = NSPredicate(format: "cacheStatus == %i", DELETING)
        let predicateSynchronized = NSPredicate(format: "cacheStatus == %i", SYNCHRONIZED)
        
        
        for entityName in entities {
            
            if let objects = CDOperation.objectsForEntity(entityName, context: context, filter: predicateNew, sort: nil) as? [NSManagedObject] {
                for object in objects {new.append(object)}
            }
            
            if let objects = CDOperation.objectsForEntity(entityName, context: context, filter: predicateInserting, sort: nil) as? [NSManagedObject] {
                for object in objects {inserting.append(object)}
            }
            
            if let objects = CDOperation.objectsForEntity(entityName, context: context, filter: predicateRelating, sort: nil) as? [NSManagedObject] {
                for object in objects {relating.append(object)}
            }
            
            if let objects = CDOperation.objectsForEntity(entityName, context: context, filter: predicateChanged, sort: nil) as? [NSManagedObject] {
                for object in objects {changed.append(object)}
            }
            
            if let objects = CDOperation.objectsForEntity(entityName, context: context, filter: predicateDeleting, sort: nil) as? [NSManagedObject] {
                for object in objects {deleting.append(object)}
            }
            
            if let objects = CDOperation.objectsForEntity(entityName, context: context, filter: predicateSynchronized, sort: nil) as? [NSManagedObject] {
                for object in objects {synchronized.append(object)}
            }
        }
        
        if new.isEmpty == false {
            NSLog("%i NEW:", new.count)
            for object in new {NSLog("  --> %@", self.friendlyObjectName(object))}
        }
        if inserting.isEmpty == false {
            NSLog("%i INSERTING:", inserting.count)
            for object in inserting {NSLog("  --> %@", self.friendlyObjectName(object))}
        }
        if relating.isEmpty == false {
            NSLog("%i RELATING:", relating.count)
            for object in relating {NSLog("  --> %@", self.friendlyObjectName(object))}
        }
        if changed.isEmpty == false {
            NSLog("%i CHANGED:", changed.count)
            for object in changed {NSLog("  --> %@", self.friendlyObjectName(object))}
        }
        if deleting.isEmpty == false {
            NSLog("%i DELETING:", deleting.count)
            for object in deleting {NSLog("  --> %@", self.friendlyObjectName(object))}
        }
        if synchronized.isEmpty == false {
            NSLog("%i SYNCHRONIZED:", synchronized.count)
            for object in synchronized {NSLog("  --> %@", self.friendlyObjectName(object))}
        }
    }
    
    // MARK: - GENERAL
    func syncMessage (message:String) {
        dispatch_async(dispatch_get_main_queue(), { () -> Void in
            NSNotificationCenter.defaultCenter().postNotificationName("syncMessage", object: message)
        })
    }
    func createRecordWithObject (object:NSManagedObject) -> CKRecord {
        let record = CKRecord(recordType: object.entity.name!)
        return updateRecord(record, withObject: object)
    }
    func updateRecord (record:CKRecord, withObject object:NSManagedObject) -> CKRecord {
        
        if let entityName = object.entity.name {
            if record.recordType == entityName {
                    
                for (attribute, description) in object.entity.attributesByName {
                    if let type = description.attributeValueClassName {
                        
                        if attribute != "cacheID" && attribute != "cacheStatus" {
                            
                            switch (type) {
                                
                            case "NSNumber":
                                if let value = object.valueForKey(attribute) as? NSNumber {
                                    record.setObject(value, forKey: attribute)
                                }
                                
                            case "NSDecimalNumber":
                                if let value = object.valueForKey(attribute) as? NSDecimalNumber {
                                    record.setObject(value, forKey: attribute)
                                }
                                
                            case "NSString":
                                if let value = object.valueForKey(attribute) as? NSString {
                                    record.setObject(value, forKey: attribute)
                                }
                                
                            case "NSDate":
                                if let value = object.valueForKey(attribute) as? NSDate {
                                    record.setObject(value, forKey: attribute)
                                }
                                
                            case "NSData":
                                if let value = object.valueForKey(attribute) as? NSData {
                                    record.setObject(value, forKey: attribute)
                                }
                                
                            default:
                                NSLog("UNHANDLED type '%@' in %@", type, __FUNCTION__)
                            }
                        }
                    }
                }
                
                record.setObject(record.recordID.recordName, forKey: "cacheID")
                record.setObject(self.group(), forKey: "group")
                record.setObject(self.groupKey(), forKey: "groupKey")
                
            } else {NSLog("SKIPPED record update because the given object entity and record type don't match in %@", __FUNCTION__)}
        } else {NSLog("SKIPPED record update because the entityName could not be determined in %@", __FUNCTION__)}
        return record
    }
    func updateReferenceForRecord (record:CKRecord, withObject object:NSManagedObject) -> CKRecord {
        
        for (relationshipName, relationship) in object.entity.relationshipsByName {
            
            if relationship.toMany {
                //NSLog("SKIPPED unsupported To-Many '%@' relationship for %@ (ManyToOne will be processed on the inverse ToOne relationship)", relationshipName, self.friendlyObjectName(object))
            } else if relationship.ordered {
                NSLog("SKIPPED Ordered '%@' relationship for %@ (Not supported)", relationshipName, self.friendlyObjectName(object))
            } else {
                
                // ToOne relationship determined, find the related object's cacheID
                if let relatedObject = object.valueForKey(relationshipName) as? NSManagedObject {
                    if let relatedObjectCacheID = relatedObject.valueForKey("cacheID") as? String {
                        
                        let relatedObjectRecordID = CKRecordID(recordName: relatedObjectCacheID)
                        let reference = CKReference(recordID: relatedObjectRecordID, action: .None) 
                        record.setObject(reference, forKey: relationshipName)
                        NSLog("ESTABLISHED a To-One '%@' reference for %@ based on relationship from %@ to %@", relationshipName, self.friendlyRecordName(record),self.friendlyObjectName(object), self.friendlyObjectName(relatedObject))
                    } else {
                        NSLog("CRITICAL ERROR getting cacheID from related object %@. ObjectID = %@", self.friendlyObjectName(relatedObject),relatedObject.objectID.description)                        
                    }
                } else {
                     NSLog("ESTABLISHED NIL %@ relationship for %@", relationshipName, self.friendlyObjectName(object))
                     record.setObject(nil, forKey: relationshipName)
                }
            }
        }
        return record
    }
    func saveMappedRecords (mappedRecords:[CKRecord:NSManagedObject], context:NSManagedObjectContext, database:CKDatabase, successStatus:Int, fault:Bool, completion:(savedMappedRecords:[CKRecord:NSManagedObject]?) -> Void) {

        // Nothing to save
        if mappedRecords.count == 0 {completion(savedMappedRecords: nil);return}
        
        // Split into batches to prevent "too many items" & "rate limited" errors
        let batchSize = 300
        let batchCountFloat:Float = Float(mappedRecords.count) / Float(batchSize)
        var batchCountInt:Int = Int(round(batchCountFloat))
        if batchCountFloat > Float(batchCountInt) {batchCountInt++} // Round Up
        if batchCountInt == 0 {completion(savedMappedRecords: nil);return}
        var batches = [[CKRecord]]()
        let allKeys:[CKRecord] = Array(mappedRecords.keys)
        var rangeStart = 0, rangeEnd = 0
        NSLog("SAVING %i record(s) in %i batch(es) with a size of %i", mappedRecords.count, batchCountInt, batchSize)
        for batchNumber in 1...batchCountInt {
            rangeStart = rangeEnd
            rangeEnd = rangeStart + batchSize
            if rangeEnd > allKeys.count {rangeEnd = allKeys.count}

            // Pull a subset of mapped records into the "batch" sized dictionary.
            let batchKeys = allKeys[rangeStart...(rangeEnd-1)]
            var batch:[CKRecord] = []
            for batchKey in batchKeys {batch.append(batchKey)}
            batches.append(batch)
            NSLog("PROCESSING save batch %i range %i to %i", batchNumber, rangeStart, rangeEnd - 1)
        }
        
        var allSavedRecords = [CKRecord]()        
        var batchesRemaining = batches.count
        for batch in batches {
           
            let operation = CKModifyRecordsOperation(recordsToSave: batch, recordIDsToDelete: [])
            operation.savePolicy = CKRecordSavePolicy.AllKeys
            operation.qualityOfService = .UserInteractive
            operation.modifyRecordsCompletionBlock = {(savedRecords, deletedRecords, error) -> Void in
                
                if let _error = error {
                    self.syncMessage("Sync Error 1. Try Later")
                    NSLog("ERROR saving records: %@", _error)
                    completion(savedMappedRecords: nil)
                } else if let _savedRecords = savedRecords {
                    // Reconstitute batch
                    allSavedRecords.appendContentsOf(_savedRecords)
                }
                batchesRemaining--
                NSLog("%i save batch(es) remaining", batchesRemaining)
                if batchesRemaining == 0 {
                    var savedMappedRecords = [CKRecord:NSManagedObject]()
                    for savedRecord in allSavedRecords {
                        if let mappedObject = mappedRecords[savedRecord] {
                            
                            // Set cacheID if it's missing
                            if mappedObject.valueForKey("cacheID") == nil {
                                mappedObject.setValue(savedRecord.recordID.recordName, forKey: "cacheID")
                            }
                            self.syncMessage("Synchronized \(self.friendlyRecordSummary(savedRecord))")
                            NSLog("SAVED record %@", self.friendlyRecordName(savedRecord))
                            
                            // Update Cache Status
                            self.setCacheStatusForObject(mappedObject, requestedStatus: successStatus)
                            savedMappedRecords[savedRecord] = mappedObject
                            
                            // Free up memory
                            if fault {CDFaulter.faultObject(mappedObject, moc: context)}
                        }
                    }
                    completion(savedMappedRecords: savedMappedRecords)
                }
            }
            database.addOperation(operation)
        }
    }
    
    // MARK: - INSERTS
    func uploadNewRecordsForEntities(entities:[String], context:NSManagedObjectContext, database:CKDatabase, completion:() -> Void) {
        
        // Prepare new records to insert
        var recordsToInsert = [CKRecord:NSManagedObject]()
        let predicate = NSPredicate(format: "cacheStatus == %i OR cacheStatus == %i", NEW, INSERTING)
        for entityName in entities {
            if let objects = CDOperation.objectsForEntity(entityName, context: context, filter: predicate, sort: nil) as? [NSManagedObject] {
                
                // Set Cache Status to INSERTING for given records
                for object in objects {
                    self.setCacheStatusForObject(object, requestedStatus: INSERTING)
                    let record = self.createRecordWithObject(object)
                    recordsToInsert[record] = object
                }
                
                // Ensure objects have a permanentID, which ensures that references to these objects via relationships work properly
                do {
                    
                    try context.obtainPermanentIDsForObjects(objects)
                    
                } catch {
                    NSLog("ERROR obtaining permanentIDsForObjects in %@",__FUNCTION__)
                }
                CDHelper.save(context)
            }
        }
        if recordsToInsert.isEmpty {NSLog("No new objects need uploading.");completion();return}
        
        self.saveMappedRecords(recordsToInsert, context: context, database: database, successStatus: RELATING, fault: false) { (savedMappedRecords) -> Void in
            
            if let _savedMappedRecords = savedMappedRecords {
                
                self.uploadReferencesForRecords(_savedMappedRecords, context: context, database: database, completion: { () -> Void in
                    completion()
                })
                
            } else {
                NSLog("FAILED to insert records, setting the cacheStatus back to New")
                for (_, object) in recordsToInsert {
                    self.setCacheStatusForObject(object, requestedStatus: NEW)
                }
                completion()
            }
        }
    }
    func uploadReferencesForRecords  (records:[CKRecord:NSManagedObject], context:NSManagedObjectContext, database:CKDatabase, completion:() -> Void) {
        
        var recordsToSave = [CKRecord:NSManagedObject]()
        
        // Upload references
        for (record, object) in records {
            let updatedRecord = self.updateReferenceForRecord(record, withObject: object)
            recordsToSave[updatedRecord] = object
        }
        
        self.saveMappedRecords(recordsToSave, context: context, database: database, successStatus: SYNCHRONIZED, fault: true) { (savedMappedRecords) -> Void in
            CDHelper.save(context)
            completion()
        }
    }
    
    // MARK: - TRACKING
    func setDateKey(key:String, date:NSDate) {
        
        let defaults = NSUserDefaults.standardUserDefaults()
        defaults.setObject(date, forKey: key)    
        NSLog("%@ has been set to %@", key, date)
    }
    func dateKey (key:String) -> NSDate {
        
        let defaults = NSUserDefaults.standardUserDefaults()    
        if let object = defaults.objectForKey(key) as? NSDate {
            return object
        }
        let object = NSDate(timeIntervalSince1970: 0) // Default
        defaults.setObject(object, forKey: key)
        return object
    } 
    
    // MARK: - SYNC
    func syncEntities (entities:[String], database:CKDatabase, context:NSManagedObjectContext) {
        
        if CDHelper.shared.iCloudAccountIsSignedIn() == false {
            NSLog("\n\nSKIPPED SYNC - iCloud Account is NOT signed in.")
            self.syncMessage("Groceries")
            return
        }
        
        // Limit Sync calls to once per 10 seconds.
        let syncLockout = dateKey("SYNC LOCKOUT")
        if syncLockout.timeIntervalSince1970 > NSDate().timeIntervalSince1970 {
            NSLog("\n\nSKIPPED SYNC - Sync was performed within the last 10 seconds")
            self.syncMessage("Groceries")
            return
        }
        self.setDateKey("SYNC LOCKOUT", date: NSDate(timeIntervalSinceNow: 10))
        
        // Sync
        context.performBlock { () -> Void in
            
            NSLog("----------------STARTED uploadNewRecordsForEntities----------------")
            self.syncMessage("Uploading New Items")
            self.uploadNewRecordsForEntities(entities, context: context, database: database, completion: { () -> Void in
                NSLog("----------------FINISHED uploadNewRecordsForEntities----------------")
                self.cacheStatusForAllObjects(entities, context: context)
                self.syncMessage("Groceries")
            })
        }
    }
}
