//
//  Item.swift
//  Groceries
//
//  Created by Tim Roadley on 29/09/2015.
//  Copyright © 2015 Tim Roadley. All rights reserved.
//

import Foundation
import CoreData

class Item: NSManagedObject {

    class func ensureLocationsAreNotNilForAllItems () {
        
        let predicate = NSPredicate(format: "locationAtHome == nil OR locationAtShop == nil")
        let lostItems = CDOperation.objectsForEntity("Item", context: CDHelper.shared.context, filter: predicate, sort: nil)
        
        if let items = lostItems as? [Item] {
            
            for item in items {
                
                // Assign orphaned items to unknown home/shop location
                ensureHomeLocationIsNotNil(item)
                ensureShopLocationIsNotNil(item)
                
                // Update item status & save                
                if let currentStatus = item.valueForKey("cacheStatus") as? NSNumber {
                    if currentStatus == SYNCHRONIZED {
                        CDCloudSync.shared.setCacheStatusForObject(item, requestedStatus: CHANGED)
                    } else {
                        if let context = item.managedObjectContext {CDHelper.save(context)}
                    }
                } else {
                    print("ERROR getting cacheStatus in \(__FUNCTION__)")
                }
            }
        }
    }
    class func ensureHomeLocationIsNotNil(item:Item) {
        
        if item.locationAtHome != nil {return}
        item.modified = NSDate()

        let context   = item.managedObjectContext!
        let entity    = "LocationAtHome"
        let attribute = "storedIn"
        let value     = "? Unknown Location"
        
        if let unknown = CDOperation.objectWithAttributeValueForEntity(entity, context:context, attribute:attribute, value:value) as? LocationAtHome {
            item.locationAtHome = unknown
            
        } else {
            if let newLocation = NSEntityDescription.insertNewObjectForEntityForName("LocationAtHome", inManagedObjectContext:context) as? LocationAtHome {
                
                print("CREATED ? Unknown Location for HOME")
                CDCloudSync.addGroupInfo(newLocation)
                newLocation.storedIn = value
                item.locationAtHome = newLocation
                newLocation.modified = NSDate()
                if let context = newLocation.managedObjectContext {CDHelper.save(context)}
            }
        }
    }
    class func ensureShopLocationIsNotNil(item:Item) {
        
        if item.locationAtShop != nil {return}
        item.modified = NSDate()

        let context   = item.managedObjectContext!
        let entity    = "LocationAtShop"
        let attribute = "aisle"
        let value     = "? Unknown Location"
        
        if let unknown = CDOperation.objectWithAttributeValueForEntity(entity, context:context, attribute:attribute, value:value) as? LocationAtShop {
            item.locationAtShop = unknown
        } else {
            if let newLocation = NSEntityDescription.insertNewObjectForEntityForName("LocationAtShop", inManagedObjectContext:context) as? LocationAtShop {
                
                print("CREATED ? Unknown Location for SHOP")
                CDCloudSync.addGroupInfo(newLocation)
                newLocation.aisle = value
                item.locationAtShop = newLocation
                newLocation.modified = NSDate()
                if let context = newLocation.managedObjectContext {CDHelper.save(context)}
            }
        }
        CDCloudSync.shared.setCacheStatusForObject(item, requestedStatus: CHANGED)
        if let context = item.managedObjectContext {CDHelper.save(context)}
    }
}
