//
//  Item.swift
//  Groceries
//
//  Created by Tim Roadley on 29/09/2015.
//  Copyright © 2015 Tim Roadley. All rights reserved.
//

import Foundation
import CoreData

class Item: NSManagedObject {

    class func ensureLocationsAreNotNilForAllItems () {
        
        let predicate = NSPredicate(format: "locationAtHome == nil OR locationAtShop == nil")
        let lostItems = CDOperation.objectsForEntity("Item", context: CDHelper.shared.context, filter: predicate, sort: nil)
        
        if let items = lostItems as? [Item] {
            
            for item in items {
                // Assign orphaned items to unknown home/shop location
                ensureHomeLocationIsNotNil(item)
                ensureShopLocationIsNotNil(item)
            }
        }
    }
    class func ensureHomeLocationIsNotNil(item:Item) {
        
        if item.locationAtHome != nil {return}
        
        let context   = item.managedObjectContext!
        let entity    = "LocationAtHome"
        let attribute = "storedIn"
        let value     = "? Unknown Location"
        
        if let unknown = CDOperation.objectWithAttributeValueForEntity(entity, context:context, attribute:attribute, value:value) as? LocationAtHome {
            item.locationAtHome = unknown
        } else {
            if let newLocation = NSEntityDescription.insertNewObjectForEntityForName("LocationAtHome", inManagedObjectContext:context) as? LocationAtHome {
                CDCloudSync.addGroupInfo(newLocation)
                newLocation.storedIn = value
                item.locationAtHome = newLocation
            }
        }
    }
    class func ensureShopLocationIsNotNil(item:Item) {
        
        if item.locationAtShop != nil {return}
        
        let context   = item.managedObjectContext!
        let entity    = "LocationAtShop"
        let attribute = "aisle"
        let value     = "? Unknown Location"
        
        if let unknown = CDOperation.objectWithAttributeValueForEntity(entity, context:context, attribute:attribute, value:value) as? LocationAtShop {
            item.locationAtShop = unknown
        } else {
            if let newLocation = NSEntityDescription.insertNewObjectForEntityForName("LocationAtShop", inManagedObjectContext:context) as? LocationAtShop {
                CDCloudSync.addGroupInfo(newLocation)
                newLocation.aisle = value
                item.locationAtShop = newLocation
            }
        }
    }
}
