//
//  LocationAtHome.swift
//  Groceries
//
//  Created by Tim Roadley on 1/10/2015.
//  Copyright © 2015 Tim Roadley. All rights reserved.
//

import Foundation
import CoreData
import CloudKit

class LocationAtHome: Location {
    
    class func printItems (locationAtHome:LocationAtHome) {
        
        if let storedIn = locationAtHome.storedIn, cacheID = locationAtHome.cacheID {
            if let items = locationAtHome.items {
                
                NSLog("The '%@' home location [%@] is related to %i items",storedIn,cacheID,items.count)
                for item in items {
                    if let name = item.valueForKey("name") as? String {
                        NSLog("  - %@", name)
                    } else {NSLog("Failed to retrieve the item name value from item")}
                }
            } else {NSLog("Failed to retrieve the items value from home location")}
        } else {NSLog("Failed to retrieve the storedIn value from home location")}
    }
    
    class func migrateItemsAndDeleteSource (source:LocationAtHome, target:LocationAtHome) {
        
        // Move item relationship from source location to target location
        if let sourceItems = source.items {
            for sourceItem in sourceItems {
                if let _sourceItem = sourceItem as? Item {
                    
                    _sourceItem.locationAtHome = target
                    CDCloudSync.shared.setCacheStatusForObject(_sourceItem, requestedStatus: CHANGED)
                }
            }
        }
        source.storedIn = source.cacheID // Prevent subsequent deduplication triggers
        source.items = nil
        LocationAtHome.printItems(target)
        CDCloudSync.shared.setCacheStatusForObject(target, requestedStatus: CHANGED)
        LocationAtHome.printItems(source)
        CDCloudSync.shared.setCacheStatusForObject(source, requestedStatus: DELETING)
        
        // Refresh UI.
        // This is required because NSFetchedResultsController doesn't track related objects.
        // The prepare and home tables are grouped by related objects.
        NSNotificationCenter.defaultCenter().postNotificationName("SomethingChanged", object: nil)
    }
    
    class func deduplicate (objectA objectA:NSManagedObject, objectB:NSManagedObject) -> Bool {
        
        if let homeA = objectA as? LocationAtHome, homeB = objectB as? LocationAtHome, storedIn = homeA.storedIn {
            
            if let homeAcacheStatus = homeA.cacheStatus {
                if homeAcacheStatus == DELETING {
                    NSLog("SKIPPED '%@'de-duplication because one of the duplicates has already been deleted",storedIn)
                    return true // deduplication attempted
                }
            }
            
            if let homeBcacheStatus = homeB.cacheStatus {
                if homeBcacheStatus == DELETING {
                    NSLog("SKIPPED '%@'de-duplication because one of the duplicates has already been deleted",storedIn)
                    return true // deduplication attempted
                }
            }
            
            NSLog("De-duplicating the '%@' Home Location", storedIn)
            if let homeAcacheID = homeA.cacheID, homeBcacheID = homeB.cacheID {
                
                if let first = [homeAcacheID,homeBcacheID].sort().first {
                    
                    if first == homeAcacheID {
                        NSLog("Removing %@ and leaving %@ in place",CDCloudSync.shared.friendlyObjectName(homeA),CDCloudSync.shared.friendlyObjectName(homeB))
                        LocationAtHome.migrateItemsAndDeleteSource(homeA, target: homeB)
                    } else {
                        NSLog("Removing %@ and leaving %@ in place",CDCloudSync.shared.friendlyObjectName(homeB),CDCloudSync.shared.friendlyObjectName(homeA))
                        LocationAtHome.migrateItemsAndDeleteSource(homeB, target: homeA)
                    }
                }
            } else {NSLog("SKIPPED de-duplication because a home location's cacheID is nil")}
            return true // deduplication attempted
        }
        return false // logic skipped
    }
}
