//
//  LocationAtShop.swift
//  Groceries
//
//  Created by Tim Roadley on 1/10/2015.
//  Copyright © 2015 Tim Roadley. All rights reserved.
//

import Foundation
import CoreData
import CloudKit

class LocationAtShop: Location {
    
    class func printItems (locationAtShop:LocationAtShop) {
        
        if let aisle = locationAtShop.aisle, cacheID = locationAtShop.cacheID {
            if let items = locationAtShop.items {
                
                NSLog("The '%@' shop location [%@] is related to %i items",aisle,cacheID,items.count)
                for item in items {
                    if let name = item.valueForKey("name") as? String {
                        NSLog("  - %@", name)
                    } else {NSLog("Failed to retrieve the item name value from item")}
                }
            } else {NSLog("Failed to retrieve the items value from shop location")}
        } else {NSLog("Failed to retrieve the aisle value from shop location")}
    }
    
    class func migrateItemsAndDeleteSource (source:LocationAtShop, target:LocationAtShop) {
        
        // Move item relationship from source location to target location
        if let sourceItems = source.items {
            for sourceItem in sourceItems {
                if let _sourceItem = sourceItem as? Item {
                    
                    _sourceItem.locationAtShop = target
                    CDCloudSync.shared.setCacheStatusForObject(_sourceItem, requestedStatus: CHANGED)
                }
            }
        }
        source.aisle = source.cacheID // Prevent subsequent deduplication triggers
        source.items = nil
        LocationAtShop.printItems(target)
        CDCloudSync.shared.setCacheStatusForObject(target, requestedStatus: CHANGED)
        LocationAtShop.printItems(source)
        CDCloudSync.shared.setCacheStatusForObject(source, requestedStatus: DELETING)
        
        // Refresh UI.
        // This is required because NSFetchedResultsController doesn't track related objects.
        // The prepare and shop tables are grouped by related objects.
        NSNotificationCenter.defaultCenter().postNotificationName("SomethingChanged", object: nil)
    }
    
    class func deduplicate (objectA objectA:NSManagedObject, objectB:NSManagedObject) -> Bool {
        
        if let shopA = objectA as? LocationAtShop, shopB = objectB as? LocationAtShop, aisle = shopA.aisle {
            
            if let shopAcacheStatus = shopA.cacheStatus {
                if shopAcacheStatus == DELETING {
                    NSLog("SKIPPED '%@'de-duplication because one of the duplicates has already been deleted",aisle)
                    return true // deduplication attempted
                }
            }
            
            if let shopBcacheStatus = shopB.cacheStatus {
                if shopBcacheStatus == DELETING {
                    NSLog("SKIPPED '%@'de-duplication because one of the duplicates has already been deleted",aisle)
                    return true // deduplication attempted
                }
            }
            
            NSLog("De-duplicating the '%@' Shop Location", aisle)
            if let shopAcacheID = shopA.cacheID, shopBcacheID = shopB.cacheID {
                
                if let first = [shopAcacheID,shopBcacheID].sort().first {
                    
                    if first == shopAcacheID {
                        NSLog("Removing %@ and leaving %@ in place",CDCloudSync.shared.friendlyObjectName(shopA),CDCloudSync.shared.friendlyObjectName(shopB))
                        LocationAtShop.migrateItemsAndDeleteSource(shopA, target: shopB)
                    } else {
                        NSLog("Removing %@ and leaving %@ in place",CDCloudSync.shared.friendlyObjectName(shopB),CDCloudSync.shared.friendlyObjectName(shopA))
                        LocationAtShop.migrateItemsAndDeleteSource(shopB, target: shopA)
                    }
                }
            } else {NSLog("SKIPPED de-duplication because a shop location's cacheID is nil")}
            return true // deduplication attempted
        }
        return false // logic skipped
    }
}
