//
//  UnitPickerTF.swift
//  Groceries
//
//  Created by Tim Roadley on 5/10/2015.
//  Copyright © 2015 Tim Roadley. All rights reserved.
//

import UIKit
import CoreData

class UnitPickerTF: CDPickerTextField {

    // MARK: - INITIALIZATION
    required init?(coder aDecoder: NSCoder) {
        super.init(coder: aDecoder)
        
        // CDPickerTextField subclass customization
        self.entity = "Unit"
        self.sort = [NSSortDescriptor(key: "name", ascending: true)]
        self.filter = NSPredicate(format: "group == %@ AND groupKey == %@ AND cacheStatus != %i", CDCloudSync.shared.group(), CDCloudSync.shared.groupKey(), DELETING)
        self.fetchBatchSize = 10
        self.performFetch()
        
        self.inputView = self.createInputView()
        self.inputAccessoryView = self.createInputAccessoryView()
        
        NSNotificationCenter.defaultCenter().addObserver(self, selector: "groupChanged", name: "GroupChanged", object: nil)
    }
    
    // MARK: - ROW SELECTION
    func setSelectedRowForComponent (component:Int) {

        switch (component) {
            case 0:
    
                switch (picker, selectedTitle, self.frc.fetchedObjects as? [Unit]) {
                    case (.Some(let _picker), .Some(let selectedTitle), .Some(let units)):
    
                        // Find the row with text matching the selected title
                        var row = 0
                        for (unit) in units {
                            if selectedTitle == unit.name {
                                _picker.selectRow(row, inComponent: component, animated: false)
                                break
                            }
                            row++
                        }
    
                    default:
                        print("_picker or selectedTitle is nil or fetchedObjects not [Unit]")
                }

                break;
                // Add more cases to support additional picker components
            default:
                break;
        }
    }

    // MARK: - DATASOURCE: UIPickerView
    override func pickerView(pickerView: UIPickerView, titleForRow row: Int, forComponent component: Int) -> String? {

        let indexPath = NSIndexPath(forRow: row, inSection: component)
        if let unit = self.frc.objectAtIndexPath(indexPath) as? Unit {
            return unit.name
        }
        return ""
    }
    
    // MARK: - CLOUDSYNC
    func groupChanged () {
        let predicate = NSPredicate(format: "group == %@ AND groupKey == %@ AND cacheStatus != %i", CDCloudSync.shared.group(), CDCloudSync.shared.groupKey(), DELETING)
        self.filter = predicate
        self.frc.fetchRequest.predicate = predicate
        self.performFetch()
    }
}
