//
//  UnitsTVC.swift
//  Groceries
//
//  Created by Tim Roadley on 4/10/2015.
//  Copyright © 2015 Tim Roadley. All rights reserved.
//

import UIKit
import CoreData

class UnitsTVC: CDTableViewController {
    
    // MARK: - CELL CONFIGURATION
    override func configureCell(cell: UITableViewCell, atIndexPath indexPath: NSIndexPath) {
    
        cell.accessoryType = UITableViewCellAccessoryType.None
    
        if let unit = self.frc.objectAtIndexPath(indexPath) as? Unit {
            if let textLabel = cell.textLabel {
                textLabel.text = unit.name
            }
        }
    }
    
    // MARK: - INITIALIZATION
    required init?(coder aDecoder: NSCoder) {
        super.init(coder: aDecoder)
        
        // CDTableViewController subclass customization
        self.entity = "Unit"
        self.sort = [NSSortDescriptor(key: "name", ascending: true)]
        self.filter = NSPredicate(format: "group == %@ AND groupKey == %@ AND cacheStatus != %i", CDCloudSync.shared.group(), CDCloudSync.shared.groupKey(), DELETING)
        self.fetchBatchSize = 25
        
        NSNotificationCenter.defaultCenter().addObserver(self, selector: "groupChanged", name: "GroupChanged", object: nil)
    }
    
    // MARK: - VIEW
    override func viewDidLoad() {
        super.viewDidLoad()
        self.navigationItem.title = "Units"
        self.performFetch()
    }
    
    // MARK: - DATA SOURCE: UITableView
    override func tableView(tableView: UITableView, commitEditingStyle editingStyle: UITableViewCellEditingStyle, forRowAtIndexPath indexPath: NSIndexPath) {
        
        if let object = self.frc.objectAtIndexPath(indexPath) as? NSManagedObject {
            CDCloudSync.shared.setCacheStatusForObject(object, requestedStatus: DELETING)
        }
        CDHelper.saveSharedContext()
    }

    // MARK: - INTERACTION
    @IBAction func done (sender: AnyObject) {
        
        if let parent = self.parentViewController {
            parent.dismissViewControllerAnimated(true, completion: nil)
        }
    }
    
    // MARK: - SEGUE
    override func prepareForSegue(segue: UIStoryboardSegue, sender: AnyObject?) {
        if let unitVC = segue.destinationViewController as? UnitVC {
                
            if segue.identifier == "Add Object Segue" {

                let object = CDOperation.insertUniqueObject("Unit", context: CDHelper.shared.context, uniqueAttributes: ["name":""], additionalAttributes: nil)
                CDCloudSync.addGroupInfo(object)
                unitVC.segueObject = object

            } else if segue.identifier == "Edit Object Segue" {
        
                if let indexPath = self.tableView.indexPathForSelectedRow {
                    if let object = self.frc.objectAtIndexPath(indexPath) as? NSManagedObject {
                        unitVC.segueObject = object
                    }
                }
            } else {print("Unidentified Segue Attempted!")}
        }
    }
    
    // MARK: - CLOUDSYNC
    func groupChanged () {
        let predicate = NSPredicate(format: "group == %@ AND groupKey == %@ AND cacheStatus != %i", CDCloudSync.shared.group(), CDCloudSync.shared.groupKey(), DELETING)
        self.filter = predicate
        self.frc.fetchRequest.predicate = predicate
        self.performFetch()
    }
}
